#ifndef SIMULATIONS_H
#define SIMULATIONS_H

#include "../Utilities/vector2D.h"
#include <num/num.h>
#include "pinball_entities.h"


extern Scene MyPinball;


void HandleBallBallCollision(Ball *ball1, Ball *ball2) {
  libnum::num32 restitution = MIN(ball1->restitution, ball2->restitution);
  Vector2D dir = ball2->pos - ball1->pos;
  libnum::num32 d = dir.Length();
  if (d == libnum::num32(0) || d > (ball1->radius + ball2->radius))
    return;

  dir.Normalise();

  libnum::num32 corr = (ball1->radius + ball2->radius - d) / libnum::num32(2);
  ball1->pos.Add(dir, -corr);
  ball2->pos.Add(dir, corr);

  libnum::num32 v1 = ball1->vel.Dot(dir);
  libnum::num32 v2 = ball2->vel.Dot(dir);

  libnum::num32 m1 = ball1->mass;
  libnum::num32 m2 = ball2->mass;

  libnum::num32 newV1 =
      (m1 * v1 + m2 * v2 - m2 * (v1 - v2) * restitution / (m1 + m2));
  libnum::num32 newV2 =
      (m1 * v1 + m2 * v2 - m1 * (v2 - v1) * restitution / (m1 + m2));

  ball1->vel.Add(dir, newV1 - v1);
  ball2->vel.Add(dir, newV2 - v2);
}

void HandleBallObstacleCollision(Ball *ball, Obstacle obstacle) {
  Vector2D dir = ball->pos - obstacle.pos;
  libnum::num32 d = dir.Length();
  if (d == libnum::num32(0) || d > (ball->radius + obstacle.radius))
    return;

  dir.Normalise();

  libnum::num32 corr = ball->radius + obstacle.radius - d;
  ball->pos.Add(dir, corr);

  libnum::num32 v = ball->vel.Dot(dir);
  ball->vel.Add(dir, obstacle.pushVel - v);

  MyPinball.score += obstacle.points;
}

void HandleBallIslandCollision(Ball *ball, std::vector<Vector2D> island,
                               libnum::num32 rad) {
  int mod = island.size();
  for (int i = 0; i < mod; i++) {
    Vector2D closest =
        ClosestPointOnSegment(ball->pos, island[i], island[(i + 1) % mod]);

    Vector2D dir;
    dir.SubtractVectors(ball->pos, closest);
    libnum::num32 d = dir.Length();
    if (!(d == libnum::num32(0) || d > (ball->radius + rad))) {
      dir.Normalise();

      libnum::num32 corr = ball->radius + rad - d;
      ball->pos.Add(dir, corr);

      /* Update velocity */

      Vector2D radius = closest.Clone();
      radius.Add(dir, rad);
      radius.Subtract(island[i], libnum::num32(1));

      libnum::num32 v = ball->vel.Dot(dir);
      libnum::num32 newV = ABS(v) * ball->restitution;
      ball->vel.Add(dir, newV - v);
    }
  }
}

void HandleBallFlipperCollision(Ball *ball, Flipper flipper) {
  Vector2D closest =
      ClosestPointOnSegment(ball->pos, flipper.pos, flipper.getTip());

  Vector2D dir;
  dir.SubtractVectors(ball->pos, closest);
  libnum::num32 d = dir.Length();
  if (d == libnum::num32(0) || d > (ball->radius + flipper.radius))
    return;

  dir.Normalise();

  libnum::num32 corr = ball->radius + flipper.radius - d;
  ball->pos.Add(dir, corr);

  /* Update velocity */

  Vector2D radius = closest.Clone();
  radius.Add(dir, flipper.radius);
  radius.Subtract(flipper.pos, libnum::num32(1));

  Vector2D surfaceVel = radius.PerpCW();
  surfaceVel.Scale(flipper.currentAngularVelocity);

  libnum::num32 v = ball->vel.Dot(dir);
  libnum::num32 newV = surfaceVel.Dot(dir);

  ball->vel.Add(dir, newV - v);
}

void HandleBallBorderCollision(Ball *ball, std::vector<Vector2D> border) {
  int mod = border.size();

  if (mod < 3)
    return;

  /* Find closest segment */

  Vector2D d, closest, ab, normal;

  libnum::num32 minDist = libnum::num32(0);

  for (int i = 0; i < mod; i++) {
    Vector2D a = border[i];
    Vector2D b = border[(i + 1) % mod];
    Vector2D c = ClosestPointOnSegment(ball->pos, a, b);
    d.SubtractVectors(ball->pos, c);
    libnum::num32 dist = d.Length();

    if (i == 0 || dist < minDist) {
      minDist = dist;
      closest.Set(c);
      ab.SubtractVectors(b, a);
      normal = ab.PerpCW();
    }
  }

  /* Push out */

  d.SubtractVectors(ball->pos, closest);
  libnum::num32 dist = d.Length();

  if (dist == libnum::num32(0)) {
    d.Set(normal);
    dist = normal.Length();
  }
  d.Normalise();

  if (d.Dot(normal) >= libnum::num32(0)) {
    if (dist > ball->radius)
      return;

    ball->pos.Add(d, ball->radius - dist);
  } else
    ball->pos.Add(d, -(ball->radius + dist));

  /* Update velocity */
  libnum::num32 v = ball->vel.Dot(d);
  libnum::num32 newV = ABS(v) * ball->restitution;

  ball->vel.Add(d, newV - v);
}


#endif